#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : pipelines.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 07.08.2021


import  os
import  re
import  time
import  numbers
import  textwrap
import  warnings
from    abc                 import  ABC, abstractmethod
from    concurrent.futures  import  ProcessPoolExecutor

import  numpy               as      np
import  pandas              as      pd
import  cma

from    tqdm                import  tqdm
from    joblib              import  Parallel, delayed

from    .iterable_samples   import  PEPTObject, IterableSamples
from    .point_data         import  PointData
from    .line_data          import  LineData
from    .voxels             import  Voxels
from    .utilities          import  check_iterable




class Transformer(ABC, PEPTObject):
    '''Base class for PEPT filters (transforming a sample into another) and
    reducers (transforming a list of samples).

    You should only need to subclass `Filter` and `Reducer` (or even, better,
    their more specialised subclasses, e.g. `LineDataFilter`).
    '''

    def __add__(self, other):
        # Allow adding (+) multiple filters / reducers into a pipeline
        return Pipeline([self, other])


    def __repr__(self):
        name = type(self).__name__
        post = max(40, 81 - len(name))

        # Select all attributes from dir(obj) that are not callable (i.e. not
        # methods) and don't start with "_" (i.e. they're private)
        docs = []
        for att in dir(self):
            memb = getattr(self, att)
            if not callable(memb) and not att.startswith("_"):
                memb_str = str(memb)
                if len(memb_str) > post:
                    memb_str = memb_str[:post - 2] + "..."
                memb_str.replace("\n", "\\n")

                docs.append(f"{att} = {memb_str}")

        # Build representation like:
        # TransformerName(attribute1 = [1 2 3], attribute2 = "hello world",
        #                 attribute3 = [2 4 5 7 8])
        if not len(docs):
            docstr = f"{name}()"
        else:
            docstr = f"{name}({docs[0]}, "

            prefix = " " * (len(name) + 1)
            curline = len(docstr)

            for i in range(1, len(docs)):
                if curline + len(docs[i]) > 80:
                    docstr = docstr + "\n" + prefix + docs[i] + ", "
                    curline = len(prefix) + len(docs[i]) + 2
                else:
                    docstr = docstr + docs[i] + ", "
                    curline = curline + len(docs[i]) + 2

            docstr = docstr[:-2] + ")"

        return docstr




class Filter(Transformer):
    '''Abstract class from which PEPT filters inherit. You only need to define
    a method `def fit_sample(self, sample)`, which processes a *single* sample.

    If you define a filter on `LineData`, you should subclass `LineDataFilter`.
    Same goes for `PointData` with `PointDataFilter`.
    '''

    @abstractmethod
    def fit_sample(self, sample):
        pass


    def fit(
        self,
        samples,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        '''Apply self.fit_sample (implemented by subclasses) according to the
        execution policy. Simply return a list of processed samples. If you
        need a reduction step (e.g. stack all processed samples), apply it
        in the subclass.
        '''

        if not hasattr(samples, "__iter__"):
            raise ValueError(textwrap.fill((
                "The input `samples` must be an iterable (e.g. a list, tuple "
                f"or LineData / PointData). Received `{type(samples)}`."
            )))

        if executor == "sequential":
            if verbose:
                samples = tqdm(samples, desc = f"{type(self).__name__} :")

            return [self.fit_sample(s) for s in samples]

        elif executor == "joblib":
            if verbose:
                samples = tqdm(samples, desc = f"{type(self).__name__} :")

            if max_workers is None:
                max_workers = os.cpu_count()

            return Parallel(n_jobs = max_workers)(
                delayed(self.fit_sample)(s) for s in samples
            )

        else:
            with executor(max_workers = max_workers) as exe:
                futures = [exe.submit(self.fit_sample, s) for s in samples]

                if verbose:
                    futures = tqdm(futures, desc = f"{type(self).__name__} :")

                return [f.result() for f in futures]




class Reducer(Transformer):
    '''Abstract class from which PEPT reducers inherit. You only need to define
    a method `def fit(self, samples)`, which processes an *iterable* of samples
    (most commonly a `LineData` or `PointData`).
    '''

    @abstractmethod
    def fit(self, samples):
        pass




class LineDataFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.LineData`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    def fit(
        self,
        line_data,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        check_iterable(LineData, line_data = line_data)
        return Filter.fit(
            self,
            line_data,
            executor,
            max_workers,
            verbose = verbose,
        )




class PointDataFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.PointData`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    def fit(
        self,
        point_data,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):

        check_iterable(PointData, point_data = point_data)
        return Filter.fit(
            self,
            point_data,
            executor,
            max_workers,
            verbose = verbose,
        )




class VoxelsFilter(Filter):
    '''An abstract class that defines a filter for samples of `pept.Voxels`.

    An implementor must define the method `def fit_sample(self, sample)`.

    A default `fit` method is provided for convenience, calling `fit_sample`
    on each sample from an iterable according to a given execution policy
    (e.g. "sequential", "joblib", or `concurrent.futures.Executor` subclasses,
    such as `ProcessPoolExecutor` or `MPIPoolExecutor`).
    '''

    def fit(
        self,
        voxels,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        check_iterable(Voxels, voxels = voxels)

        return Filter.fit(
            self,
            voxels,
            executor,
            max_workers,
            verbose = verbose,
        )




class Pipeline(PEPTObject):
    '''A PEPT processing pipeline, chaining multiple `Filter` and `Reducer`
    for efficient, parallel execution.

    After a pipeline is constructed, the `fit(samples)` method can be called,
    which will apply the chain of filters and reducers on the samples of data.

    A filter is simply a transformation applied to a sample (e.g. `Voxelliser`
    on a single sample of `LineData`). A reducer is a transformation applied to
    a list of *all* samples (e.g. `Stack` on all samples of `PointData`).

    Note that only filters can be applied in parallel, but the great advantage
    of a `Pipeline` is that it significantly reduces the amount of data copying
    and intermediate results' storage. Reducers will require collecting all
    results.

    There are three execution policies at the moment: "sequential" is
    single-threaded (slower, but easy to debug), "joblib" (very fast on medium
    datasets due to joblib's caching) and any `concurrent.futures.Executor`
    subclass (e.g. MPIPoolExecutor for parallel processing on distributed
    clusters).

    Attributes
    ----------
    transformers : list[pept.base.Filter or pept.base.Reducer]
        A list of transformers that will be applied consecutively to samples of
        data. A `pept.base.Filter` subclass transforms individual samples of
        data (and defines a `fit_sample` method), while a `pept.base.Reducer`
        subclass transforms entire lists of samples (and defines a `fit`
        method).

    Examples
    --------
    A pipeline can be created in two ways: either by adding (+) multiple
    transformers together, or explicitly constructing the `Pipeline` class.

    The first method is the most straightforward:

    >>> import pept

    >>> filter1 = pept.tracking.Cutpoints(max_distance = 0.5)
    >>> filter2 = pept.tracking.HDBSCAN(true_fraction = 0.1)
    >>> reducer = pept.tracking.Stack()
    >>> pipeline = filter1 + filter2 + reducer

    >>> print(pipeline)
    Pipeline
    --------
    transformers = [
        Cutpoints(append_indices = False, cutoffs = None, max_distance = 0.5)
        HDBSCAN(clusterer = HDBSCAN(), max_tracers = 1, true_fraction = 0.1)
        Stack(overlap = None, sample_size = None)
    ]

    >>> lors = pept.LineData(...)        # Some samples of lines
    >>> points = pipeline.fit(lors)

    The chain of filters can also be applied to a single sample:

    >>> point = pipeline.fit_sample(lors[0])

    The pipeline's `fit` method allows specifying an execution policy:

    >>> points = pipeline.fit(lors, executor = "sequential")
    >>> points = pipeline.fit(lors, executor = "joblib")

    >>> from mpi4py.futures import MPIPoolExecutor
    >>> points = pipeline.fit(lors, executor = MPIPoolExecutor)

    The `pept.Pipeline` constructor can also be called directly, which allows
    the enumeration of filters:

    >>> pipeline = pept.Pipeline([filter1, filter2, reducer])

    Adding new filters is very easy:

    >>> pipeline_extra = pipeline + filter2

    '''

    def __init__(self, transformers):
        '''Construct the class from an iterable of ``Filter``, ``Reducer``
        and/or other ``Pipeline`` instances (which will be flattened).
        '''

        # Type-checking inputs
        transformers = list(transformers)

        # The list of transformers might contain entire pipelines; extract
        # their filters into a flat list of transformers
        flattened = []
        for i, t in enumerate(transformers):
            if isinstance(t, Transformer):
                flattened.append(t)

            elif isinstance(t, Pipeline):
                flattened.extend(t.transformers)

            else:
                raise TypeError(textwrap.fill((
                    "All input `transformers` must be subclasses of "
                    "`pept.base.Filter`, `pept.base.Reducer` or "
                    f"`pept.base.Pipeline`. Received `{type(t)}` at index {i}."
                )))

        # Setting class attributes
        self._transformers = flattened


    @property
    def filters(self):
        '''Only the `Filter` instances from the `transformers`. They can be
        applied in parallel.
        '''
        return [t for t in self.transformers if isinstance(t, Filter)]


    @property
    def reducers(self):
        '''Only the `Reducer` instances from the `transformers`. They require
        collecting all parallel results.
        '''
        return [t for t in self.transformers if isinstance(t, Reducer)]


    @property
    def transformers(self):
        '''The list of `Transformer` to be applied; this includes both `Filter`
        and `Reducer` instances.
        '''
        return self._transformers


    def _fit_sample(self, transformers, sample):
        # It is more efficient to fit samples procedurally; recursion in Python
        # is slow and doesn't allow garbage-collecting intermediate states
        for t in transformers:
            # If we need to apply a reducer to a single sample, use [sample]
            if isinstance(t, Reducer):
                sample = t.fit([sample])
            else:
                sample = t.fit_sample(sample)

        return sample


    def fit_sample(self, sample):
        '''Apply all transformers - consecutively - to a single sample of data.
        The output type is simply what the transformers return.
        '''
        return self._fit_sample(self.transformers, sample)


    def _fit(self, transformers, samples, executor, max_workers, desc = None):
        # Select type of execution of the filtering steps and extract a list
        # of processed samples
        if executor == "sequential":
            if desc is not None:
                samples = tqdm(samples, desc = desc)

            return [self._fit_sample(transformers, s) for s in samples]

        elif executor == "joblib":
            if desc is not None:
                samples = tqdm(samples, desc = desc)

            # Joblib's `max_workers` behaviour is different than for Executors.
            # The latter simply sets the maximum number of threads available,
            # while the former uses a max_worker=1. Make behaviour consistent.
            if max_workers is None:
                max_workers = os.cpu_count()

            return Parallel(n_jobs = max_workers)(
                delayed(self._fit_sample)(transformers, s) for s in samples
            )

        else:
            # Otherwise assume `executor` is a `concurrent.futures.Executor`
            # subclass (e.g. ProcessPoolExecutor, MPIPoolExecutor).
            with executor(max_workers = max_workers) as exe:
                futures = [
                    exe.submit(self._fit_sample, transformers, s)
                    for s in samples
                ]

                if desc is not None:
                    futures = tqdm(futures, desc = desc)

                return [f.result() for f in futures]


    def fit(
        self,
        samples,
        executor = "joblib",
        max_workers = None,
        verbose = True,
    ):
        '''Apply all transformers defined to all `samples`. Filters are applied
        according to the `executor` policy (e.g. parallel via "joblib"), while
        reducers are applied on a single thread.

        Parameters
        ----------
        samples : Iterable
            An iterable (e.g. list, tuple, LineData, list[PointData]), whose
            elements will be passed through the pipeline.

        executor : "sequential", "joblib", or `concurrent.futures.Executor` \
                subclass, default "joblib"
            The execution policy controlling how the chain of filters are
            applied to each sample in `samples`; "sequential" is single
            threaded (slow, but easy to debug), "joblib" is multi-threaded
            (very fast due to joblib's caching). Alternatively, a
            `concurrent.futures.Executor` subclass can be used (e.g.
            `MPIPoolExecutor` for distributed computing on clusters).

        max_workers : int, optional
            The maximum number of workers to use for parallel executors. If
            `None` (default), the maximum number of CPUs are used.

        verbose : bool, default True
            If True, show extra information during processing, e.g. loading
            bars.
        '''

        if not hasattr(samples, "__iter__"):
            raise ValueError(textwrap.fill((
                "The input `samples` must be an iterable (e.g. list, tuple, "
                f"LineData, list[PointData]). Received `{type(samples)}`."
            )))

        # If verbose, time operation
        if verbose:
            start = time.time()

        # Aggregate processing steps into a list where consecutive filters are
        # collapsed into tuples
        steps = self.steps()
        if verbose:
            nbatches = sum((isinstance(step, tuple) for step in steps))

        nbatch = 0
        for step in steps:
            # Apply continuous sequence of filters (i.e. tuple)
            if isinstance(step, tuple):
                nbatch += 1
                desc = f"Batch {nbatch} / {nbatches} :" if verbose else None
                samples = self._fit(step, samples, executor, max_workers, desc)

            # Apply reducer on all samples in `state`
            else:
                samples = step.fit(samples)

        if verbose:
            end = time.time()
            print(f"\nProcessed samples in {end - start} s\n")

        return samples


    def steps(self):
        '''Return the order of processing steps to apply as a list where all
        consecutive sequences of filters are collapsed into tuples.

        E.g. [F, F, R, F, R, R, F, F, F] -> [(F, F), R, (F), R, R, (F, F, F)].
        '''
        start_filter = 0
        collecting_filters = False

        order = []

        for i, t in enumerate(self.transformers):
            if isinstance(t, Filter) and not collecting_filters:
                # Found the first filter in a new sequence of filters
                collecting_filters = True
                start_filter = i

            elif isinstance(t, Reducer):
                # Reducers are appended unchanged, unless we were collecting a
                # sequence of filters - in which case append filter sequence
                if collecting_filters:
                    filters_seq = tuple(self.transformers[start_filter:i])
                    order.append(filters_seq)

                    collecting_filters = False

                # Append reducer
                order.append(t)

        if collecting_filters:
            # Last running sequence of filters
            filters_seq = tuple(self.transformers[start_filter:])
            order.append(filters_seq)

        return order


    def optimise(
        self,
        lines,
        # group_by = "label",
        # error = ["frequency", "error"],
        max_evals = 200,
        executor = "joblib",
        max_workers = None,
        verbose = True,
        **free_parameters,
    ):

        # Type-checking
        if not isinstance(lines, LineData):
            lines = LineData(lines)

        if not len(free_parameters):
            raise ValueError(textwrap.fill((
                "Must define at least one free parameter to optimise, given "
                "as `param_name = [range_min, range_max]`."
            )))

        # The free parameters for each transformer - prepend the input `lors` s
        # sample_size and overlap
        # This is a list[tuple[transformer, set[parameter_name]]]
        pipe_parameters = parameters_list(self)

        # Find transformer free parameters
        # This is a list[OptParam] saving the transformer and param name
        opt_parameters = [
            OptParam(pipe_parameters, param, bounds)
            for param, bounds in free_parameters.items()
        ]

        if verbose:
            _print_pipe_params(opt_parameters)

        bounds = np.array([op.bounds for op in opt_parameters])
        scaling = 0.4 * (bounds[:, 1] - bounds[:, 0])
        bounds_scaled = bounds / scaling[:, None]

        x0 = 0.5 * (bounds[:, 0] + bounds[:, 1]) / scaling
        sigma0 = 1.

        es = cma.CMAEvolutionStrategy(x0, sigma0, dict(
            bounds = [bounds_scaled[:, 0], bounds_scaled[:, 1]],
            tolflatfitness = 10,
            maxfevals = int(max_evals),
            verbose = 3 if verbose else -9,
        ))

        # Store optimisation pipeline parameters
        if max_workers is None:
            max_workers = os.cpu_count()

        with ProcessPoolExecutor(max_workers = max_workers) as executor:
            popt = OptPipeline(self, opt_parameters, lines, executor)

            # Save optimisation evolution
            history = []

            with warnings.catch_warnings():
                warnings.simplefilter("ignore", RuntimeWarning)

                while not es.stop():
                    solutions = es.ask()
                    results = _try_solutions(solutions * scaling, popt)
                    es.tell(solutions, results[:, -1])

                    _print_after_eval(
                        es,
                        opt_parameters,
                        solutions * scaling,
                        results,
                    )
                    history.append(np.c_[solutions * scaling, results])

                    if es.sigma < 0.1:
                        break

        print(es.result)

        best_solution = es.result.xfavorite * scaling
        for op, sol in zip(popt.parameters, best_solution):
            if isinstance(op.default, numbers.Integral):
                sol = int(sol)

            setattr(op.trans, op.param, sol)


        return np.vstack(history)



    def __add__(self, other):
        if isinstance(other, Transformer):
            return Pipeline(self.transformers + [other])
        elif isinstance(other, Pipeline):
            return Pipeline(self.transformers + other.transformers)
        else:
            raise TypeError(textwrap.fill((
                "All new transformers added to the pipeline must be "
                "subclasses of `pept.base.Filter`, `pept.base.Reducer` or "
                f"`pept.base.Pipeline`. Received `{type(other)}`."
            )))


    def __repr__(self):
        name = type(self).__name__
        underline = "-" * len(name)
        return f"{name}\n{underline}\ntransformers = [\n" + textwrap.indent(
            "\n".join((t.__str__() for t in self.transformers)), '    '
        ) + "\n]"




class OptParam:
    '''Class storing a single pipeline optimisation free parameter, including
    its parent transformer (`trans`), name (`param`), range (`bounds`) and
    initial value (`default`).
    '''
    # index: int
    # trans: PEPTObject
    # param: str
    # bounds: np.ndarray
    # default: float

    def __init__(self, pipe_parameters, param, bounds):

        # Type-checking
        param = str(param)
        bounds = np.asarray(bounds, dtype = float)
        if bounds.ndim != 1 or len(bounds) != 2:
            raise ValueError(textwrap.fill((
                f"The parameter range given for `{param}` is incorrect. "
                "It must be a list formatted as [min, max]. Received "
                f"`{bounds}`."
            )))

        # Parameter name may have a number appended, e.g. "sample_size2" means
        # the second pipeline free parameter named `sample_size`
        match = re.search(r"(\d+)$", param)
        if match is None:
            pnum = 1                            # Parameter number
        else:
            pnum = int(match.group())
            param = param[:match.span()[0]]     # Remove parameter number

        self.param = param
        self.bounds = bounds

        # Find index of transformer containing the `param` free parameter
        pcur = 1
        for i, (trans, param_names) in enumerate(pipe_parameters):
            if param in param_names:
                if pcur == pnum:
                    self.index = i
                    self.trans = trans
                    self.default = getattr(trans, param)
                    break
                else:
                    pcur += 1
        else:
            raise ValueError(textwrap.fill((
                f"No free parameter named `{param}` was found in the pipeline "
                f"while looking for the N={pnum} parameter with this name."
            )))




class OptPipeline:
    '''Class storing a single pipeline optimisation free parameter, including
    its parent transformer (`trans`), name (`param`), range (`bounds`) and
    initial value (`default`).
    '''

    def __init__(
        self,
        pipeline: Pipeline,
        parameters: OptParam,
        lines: LineData,
        executor: ProcessPoolExecutor,
    ):
        self.pipeline = pipeline
        self.parameters = parameters
        self.lines = lines
        self.executor = executor




def parameters_list(pipeline: Pipeline):
    '''Return a list[tuple[TransformerName, set[str]]], where for each
    transformer in a pipeline the corresponding set contains its free parameter
    names.
    '''
    parameters = []
    for trans in pipeline.transformers:
        # Save the names of the transformer's attributes that are not hidden
        # (i.e. start with "_") and are not callable (i.e. not methods)
        trans_params = (trans, set())
        for att in dir(trans):
            if not att.startswith("_") and not callable(getattr(trans, att)):
                trans_params[1].add(att)

        parameters.append(trans_params)

    return parameters


def _print_pipe_params(opt_parameters):
    trans_strs = [
        f"{op.trans.__class__.__name__}.{op.param}"
        for op in opt_parameters
    ]
    max_len = max([len(ts) for ts in trans_strs])

    print("\nOptimising Pipeline Parameters:")
    for trans, op in zip(trans_strs, opt_parameters):
        trans = trans + (max_len - len(trans)) * ' '
        print(f"  {op.index} : {trans} : {op.bounds}")
    print()


def _try_solution(pipeline, lines):
    start = time.time()
    try:
        out = pipeline.fit(
            lines,
            executor = "sequential",
            verbose = False,
        )
    except ValueError:
        out = PointData(
            np.empty((0, 5)),
            columns = ["t", "x", "y", "z", "error"],
        )
    end = time.time()
    print(f"Pipe eval: {end - start}")

    print(out)

    # Try to keep the frequency of points ~ freq LoRs / 100
    # Calculate the difference of frequency of points to lines / 100
    times = lines["t"]
    dt = times[-1] - times[0]
    pfreq = len(out.points) / dt
    lfreq = len(lines.lines) / dt

    freq_err = (pfreq - lfreq / 100) ** 2 if pfreq else np.inf

    # Error in particle positions
    spatial_err = np.nanmedian(out["error"]) if pfreq else np.inf

    return [freq_err, spatial_err, freq_err * spatial_err]



def _try_solutions(solutions, popt: OptPipeline):

    # Create pipeline copies holding the different parameter combinations
    pipes = [popt.pipeline.copy() for _ in range(len(solutions))]

    for sol_set in solutions:
        # Set transformers' parameters to the values in `sol_set`
        for i, op in enumerate(popt.parameters):
            sol = sol_set[i]
            if isinstance(op.default, numbers.Integral):
                sol = int(sol)

            setattr(pipes[i].transformers[op.index], op.param, sol)

    # Execute the different pipelines in parallel
    futures = [
        popt.executor.submit(_try_solution, pipe, popt.lines)
        for pipe in pipes
    ]
    results = [f.result() for f in futures]
    del futures

    # Reset pipeline parameters to their default values
    # for op in popt.parameters:
    #     setattr(op.trans, op.param, op.default)

    return np.array(results)


def _print_after_eval(es, parameters, solutions, results):
    # Display evaluation results: solutions, error values, etc.
    cols = [p.param for p in parameters] + ["freq_err", "spatial_err", "error"]
    sols_results = np.c_[solutions, results]

    # Store solutions and results in a DataFrame for easy pretty printing
    sols_results = pd.DataFrame(
        data = sols_results,
        columns = cols,
        index = None,
    )

    # Display all the DataFrame columns and rows
    old_max_columns = pd.get_option("display.max_columns")
    old_max_rows = pd.get_option("display.max_rows")

    pd.set_option("display.max_columns", None)
    pd.set_option("display.max_rows", None)

    print((
        f"{sols_results}\n"
        f"    Overall Convergence: {es.sigma}\n"
        f"  Parameter Convergence: {es.result.stds}\n"
        f"   Function evaluations: {es.result.evaluations}\n---"
    ), flush = True)

    pd.set_option("display.max_columns", old_max_columns)
    pd.set_option("display.max_rows", old_max_rows)



#!/usr/bin/env python3
# -*- coding: utf-8 -*-


#    pept is a Python library that unifies Positron Emission Particle
#    Tracking (PEPT) research, including tracking, simulation, data analysis
#    and visualisation tools.
#
#    If you used this codebase or any software making use of it in a scientific
#    publication, you must cite the following paper:
#        Nicuşan AL, Windows-Yule CR. Positron emission particle tracking
#        using machine learning. Review of Scientific Instruments.
#        2020 Jan 1;91(1):013329.
#        https://doi.org/10.1063/1.5129251
#
#    Copyright (C) 2019-2021 the pept developers
#
#    This program is free software: you can redistribute it and/or modify
#    it under the terms of the GNU General Public License as published by
#    the Free Software Foundation, either version 3 of the License, or
#    (at your option) any later version.
#
#    This program is distributed in the hope that it will be useful,
#    but WITHOUT ANY WARRANTY; without even the implied warranty of
#    MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
#    GNU General Public License for more details.
#
#    You should have received a copy of the GNU General Public License
#    along with this program.  If not, see <https://www.gnu.org/licenses/>.


# File   : trajectory_separation.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 23.08.2019


import  os
import  warnings

import  numpy                       as      np
from    scipy.spatial               import  cKDTree
from    scipy.sparse.csgraph        import  minimum_spanning_tree

import  pept
import  hdbscan

from    .distance_matrix_reachable  import  distance_matrix_reachable




def trajectory_errors(
    true_positions,
    tracked_positions,
    averages = True,
    stds = False,
    errors = False,
    max_workers = None
):

    # true_positions, tracked_positions cols: [t, x, y, z, ...]
    true_positions = np.asarray(true_positions, dtype = float)
    tracked_positions = np.asarray(tracked_positions, dtype = float)

    # Construct k-d tree for fast nearest neighbour lookup.
    tree = cKDTree(true_positions[:, 1:4])

    # errors cols: [err, err_x, err_y, err_z]
    errors = np.zeros((len(tracked_positions), 4))

    if max_workers is None:
        max_workers = os.cpu_count()

    # Find closest true point to each tracked point.
    for i, pos in enumerate(tracked_positions[:, 1:4]):
        dist, index = tree.query(pos, k = 1, n_jobs = max_workers)

        errors[i, 0] = np.linalg.norm(pos - true_positions[index])
        errors[i, 1:4] = np.abs(pos - true_positions[index, 0:3])
        # errors[i, 1] = np.abs(pos[0] - true_positions[index, 0])
        # errors[i, 2] = np.abs(pos[1] - true_positions[index, 1])
        # errors[i, 3] = np.abs(pos[2] - true_positions[index, 2])

    # Append and return the results that were asked for.
    results = []

    if averages:
        results.append(np.average(errors, axis = 0))
    if stds:
        results.append(np.std(errors, axis = 0))
    if errors:
        results.append(errors)

    # If a single result was asked for, return it directly; otherwise, use a
    # tuple.
    if len(results) == 1:
        return results[0]
    else:
        return tuple(results)




class Segregate(pept.base.Reducer):
    '''Segregate the intertwined points from multiple trajectories into
    individual paths.

    Reducer signature:

    ::

              pept.PointData -> Segregate.fit -> pept.PointData
        list[pept.PointData] -> Segregate.fit -> pept.PointData
               numpy.ndarray -> Segregate.fit -> pept.PointData

    The points in `point_data` (a numpy array or `pept.PointData`) are used
    to construct a minimum spanning tree in which every point can only be
    connected to `points_window` points around it - this "window" refers to
    the points in the initial data array, sorted based on the time column;
    therefore, only points within a certain timeframe can be connected. All
    edges (or "connections") in the minimum spanning tree that are larger
    than `trajectory_cut_distance` are removed (or "cut") and the remaining
    connected "clusters" are deemed individual trajectories if they contain
    more than `min_trajectory_size` points.

    The trajectory indices (or labels) are appended to `point_data`. That
    is, for each data point (i.e. row) in `point_data`, a label will be
    appended starting from 0 for the corresponding trajectory; a label of
    -1 represents noise. If `point_data` is a numpy array, a new numpy
    array is returned; if it is a `pept.PointData` instance, a new instance
    is returned.

    This function uses single linkage clustering with a custom metric for
    spatio-temporal data to segregate trajectory points. The single linkage
    clustering was optimised for this use-case: points are only connected
    if they are within a certain `points_window` in the time-sorted input
    array. Sparse matrices are also used for minimising the memory
    footprint.

    Attributes
    ----------
    window : int
        Two points are "reachable" (i.e. they can be connected) if and only
        if they are within `points_window` in the time-sorted input
        `point_data`. As the points from different trajectories are
        intertwined (e.g. for two tracers A and B, the `point_data` array
        might have two entries for A, followed by three entries for B, then
        one entry for A, etc.), this should optimally be the largest number
        of points in the input array between two consecutive points on the
        same trajectory. If `points_window` is too small, all points in the
        dataset will be unreachable. Naturally, a larger `time_window`
        correponds to more pairs needing to be checked (and the function
        will take a longer to complete).

    cut_distance : float
        Once all the closest points are connected (i.e. the minimum
        spanning tree is constructed), separate all trajectories that are
        further apart than `trajectory_cut_distance`.

    min_trajectory_size : float, default 5
        After the trajectories have been cut, declare all trajectories with
        fewer points than `min_trajectory_size` as noise.

    See Also
    --------
    Reconnet : Connect segregated trajectories based on tracer signatures.
    PlotlyGrapher : Easy, publication-ready plotting of PEPT-oriented data.

    Examples
    --------
    A typical workflow would involve transforming LoRs into points using some
    tracking algorithm. These points include all tracers moving through the
    system, being intertwined (e.g. for two tracers A and B, the `point_data`
    array might have two entries for A, followed by three entries for B, then
    one entry for A, etc.). They can be segregated based on position alone
    using this function; take for example two tracers that go downwards (below,
    'x' is the position, and in parens is the array index at which that point
    is found).

    ::

        `points`, numpy.ndarray, shape (10, 4), columns [time, x, y, z]:
            x (1)                       x (2)
             x (3)                     x (4)
               x (5)                 x (7)
               x (6)                x (9)
              x (8)                 x (10)

    >>> import pept.tracking.trajectory_separation as tsp
    >>> points_window = 10
    >>> trajectory_cut_distance = 15    # mm
    >>> segregated_trajectories = tsp.segregate_trajectories(
    >>>     points, points_window, trajectory_cut_distance
    >>> )

    ::

        `segregated_trajectories`, numpy.ndarray, shape (10, 5),
        columns [time, x, y, z, trajectory_label]:
            x (1, label = 0)            x (2, label = 1)
             x (3, label = 0)          x (4, label = 1)
               x (5, label = 0)      x (7, label = 1)
               x (6, label = 0)     x (9, label = 1)
              x (8, label = 0)      x (10, label = 1)
    '''

    def __init__(
        self,
        window,
        cut_distance,
        min_trajectory_size = 5,
    ):
        self.window = int(window)
        self.cut_distance = float(cut_distance)
        self.min_trajectory_size = int(min_trajectory_size)


    def fit(self, points):
        # Stack the input points into a single PointData
        if not isinstance(points, pept.PointData):
            points = pept.PointData(points)

        if len(points.points) == 0:
            return points.copy(
                data = points.points[0:0],
                columns = points.columns + ["label"],
            )

        pts = points.points

        # Sort pts based on the time column (col 0) and create a C-ordered copy
        # to send to Cython.
        pts = np.asarray(pts[pts[:, 0].argsort()], dtype = float, order = "C")

        # Calculate the sparse distance matrix between reachable points. This
        # is an optimised Cython function returning a sparse CSR matrix.
        distance_matrix = distance_matrix_reachable(pts, self.window)

        # Construct the minimum spanning tree from the sparse distance matrix.
        # Note that `mst` is also a sparse CSR matrix.
        mst = minimum_spanning_tree(distance_matrix)

        # Get the minimum spanning tree edges into the [vertex 1, vertex 2,
        # edge distance] format, then sort it based on the edge distance.
        mst = mst.tocoo()
        mst_edges = np.vstack((mst.row, mst.col, mst.data)).T
        mst_edges = mst_edges[mst_edges[:, 2].argsort()]

        # Ignore deprecation warning from HDBSCAN's use of `np.bool`
        with warnings.catch_warnings():
            warnings.simplefilter("ignore", category = DeprecationWarning)

            # Create the single linkage tree from the minimum spanning tree
            # edges using internal hdbscan methods (because they're damn fast).
            # This should be a fairly quick step.
            linkage_tree = hdbscan._hdbscan_linkage.label(mst_edges)
            linkage_tree = hdbscan.plots.SingleLinkageTree(linkage_tree)

            # Cut the single linkage tree at `trajectory_cut_distance` and get
            # the cluster labels, setting clusters smaller than
            # `min_trajectory_size` to -1 (i.e. noise).
            labels = linkage_tree.get_clusters(
                self.cut_distance,
                self.min_trajectory_size,
            )

        # Append the labels to `pts`.
        return points.copy(
            data = np.c_[pts, labels],
            columns = points.columns + ["label"],
        )




class Reconnect(pept.base.Reducer):
    '''Best-fit trajectory segment reconstruction based on time, distance and
    arbitrary tracer signatures.

    Reducer signature:

    ::

              pept.PointData -> Segregate.fit -> pept.PointData
        list[pept.PointData] -> Segregate.fit -> pept.PointData
               numpy.ndarray -> Segregate.fit -> pept.PointData

    After a trajectory segregation step (e.g. using ``Segregate``), you may be
    left with multiple smaller trajectory segments. Some trajectories can be
    reconstructed even when losing the tracers for a bit.

    When a tracer is lost for less than `tmax` time and `dmax` distance, its
    trajectory segments are reconnected; if multiple condidates are possible,
    the best fit is used.

    Multiple tracer signatures can be used to improve the reconnection step;
    supply them as data column names and difference thresholds, e.g. an extra
    keyword argument ``v = 1`` will join trajectories whose difference in
    velocity is smaller than 1 m/s.

    The last `num_points` points on a segment are averaged before they are
    connected with the first `num_points` on another segment.

    *New in pept-0.4.2*

    Examples
    --------
    Reconnect segments that are closer than 1 second in time and 0.1 m apart:

    >>> from pept.tracking import *
    >>> trajectories = Reconnect(tmax = 1000, dmax = 100).fit(segments)

    You can use the `cluster_size` (set by the ``Centroids`` filter) as a
    tracer signature; allow segments to be reconnected if the difference in
    their cluster size is < 100:

    >>> trajectories = Reconnect(1000, 100, cluster_size = 100).fit(segments)

    And a velocity `v` difference < 0.1:

    >>> Reconnect(1000, 100, cluster_size = 100, v = 0.1).fit(segments)

    '''

    def __init__(
        self,
        tmax,
        dmax,
        column = "label",
        num_points = 10,
        **signatures,
    ):
        self.column = column
        self.num_points = int(num_points)

        self.tmax = float(tmax)
        self.dmax = float(dmax)

        self.signatures = dict()
        for sig, thresh in signatures.items():
            self.signatures[sig] = float(thresh)


    def fit(self, points):
        points = pept.tracking.Stack().fit(points)
        if not isinstance(points, pept.PointData):
            points = pept.PointData(points)

        # Columns corresponding to the signatures
        sig_cols = [points.columns.index(sn) for sn in self.signatures.keys()]

        trajs = pept.tracking.SplitAll(self.column).fit(points)
        trajs.sort(key = lambda traj: traj["t"][0])

        # List of connections to do, list[tuple[int, int]]
        connections = []

        # Try to forward-connect the end of trajs[i] to the start of trajs[j]
        start_times = np.array([t["t"][0] for t in trajs])

        for i in range(len(trajs)):
            # Select all future trajectories whose start time is within tmax
            cur_traj = trajs[i]
            indices = np.argwhere(
                (start_times > cur_traj["t"][-1]) &
                (start_times - cur_traj["t"][-1] < self.tmax),
            ).flatten()

            # If no feasible times were found, carry on
            if not indices.any():
                continue

            # Compute connection costs between trajectory ends
            costs = []
            for j in indices:
                e2 = trajs[i].points[-self.num_points:].mean(axis = 0)
                e1 = trajs[j].points[:self.num_points].mean(axis = 0)

                # The first cost is the distance between traj ends; the rest
                # are the signature differences
                cost = [np.linalg.norm(e2[1:4] - e1[1:4])]
                for sc in sig_cols:
                    cost.append(np.abs(e2[sc] - e1[sc]))

                costs.append(cost)

            # Keep track of trajectory indices and associated costs
            costs = np.c_[indices, np.array(costs)]

            # Remove condidate connections that have costs larger than threshs
            selection = costs[:, 1] < self.dmax
            for i, sthresh in enumerate(self.signatures.values()):
                selection = selection & (costs[:, 2 + i] < sthresh)

            costs = costs[selection]

            # If no feasible connection was found, carry on
            if not len(costs):
                continue

            # Otherwise, establish connection with minimum overall cost
            best = costs[:, 1:].mean(axis = 1).argmin()
            connection_index = int(costs[best, 0])
            connections.append((i, connection_index))

        # Set connected labels
        if isinstance(self.column, str):
            label_col = points.columns.index(self.column)
        else:
            label_col = self.column

        for i1, i2 in connections:
            trajs[i2].points[:, label_col] = trajs[i1].points[0, label_col]

        # Stack trajectories and map labels from [0, 2, 2, 3, 0] to
        # [0, 1, 1, 2, 0]
        trajs = pept.tracking.Stack().fit(trajs)

        labels = trajs.points[:, label_col]
        _, ordered = np.unique(labels, return_inverse = True)
        trajs.points[:, label_col] = ordered

        return trajs



#!/usr/bin/env python3
# -*- coding: utf-8 -*-
# File   : grids.py
# License: GNU v3.0
# Author : Andrei Leonard Nicusan <a.l.nicusan@bham.ac.uk>
# Date   : 19.11.2021


import  time
import  textwrap

import  numpy           as      np
import  konigcell       as      kc

from    pept.base       import  Reducer
from    pept            import  PointData, Voxels, Pixels
from    pept.tracking   import  Stack

# PyVista plotting is optional
try:
    import  pyvista     as      pv
except ImportError:
    pass

import  plotly.figure_factory as ff




class DynamicProbability2D(Reducer):
    '''Compute the 2D probability distribution of some tracer quantity (eg
    velocity in each cell).

    Reducer signature:

    ::

              PointData -> DynamicProbability2D.fit -> Pixels
        list[PointData] -> DynamicProbability2D.fit -> Pixels
          numpy.ndarray -> DynamicProbability2D.fit -> Pixels

    This reducer calculates the average value of the tracer quantity in each
    cell of a 2D pixel grid; it uses the full projected tracer area for the
    pixelization step, so the distribution is accurate for arbitrarily fine
    resolutions.

    Parameters
    ----------
    diameter : float
        The diameter of the imaged tracer.

    column : str or int
        The `PointData` column used to compute the probability distribution,
        given as a name (`str`) or index (`int`).

    dimensions : str or list[int], default "xy"
        The tracer coordinates used to rasterize its trajectory, given as a
        string (e.g. "xy" projects the points onto the XY plane) or a list with
        two column indices (e.g. [1, 3] for XZ).

    resolution : tuple[int, int], default (512, 512)
        The number of pixels used for the rasterization grid in the X and Y
        dimensions.

    xlim : tuple[float, float], optional
        The physical limits in the X dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    ylim : tuple[float, float], optional
        The physical limits in the y dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    max_workers : int, optional
        The maximum number of workers (threads, processes or ranks) to use by
        the parallel executor; if 1, it is sequential (and produces the
        clearest error messages should they happen). If unset, the
        ``os.cpu_count()`` is used.

    verbose : bool or str default True
        If True, time the computation and print the state of the execution.

    Examples
    --------
    Compute the velocity probability distribution of a single tracer trajectory
    having a column named "v" corresponding to the tracer velocity:

    >>> trajectories = pept.load(...)
    >>> pixels_vel = DynamicProbability2D(1.2, "v", "xy").fit(trajectories)

    Plot the pixel grid:

    >>> from pept.plots import PlotlyGrapher2D
    >>> PlotlyGrapher2D().add_pixels(pixels_vel).show()

    For multiple tracer trajectories, you can use ``Segregate`` then
    ``SplitAll('label')`` before calling this reducer to rasterize each
    trajectory separately:

    >>> vel_pipeline = pept.Pipeline([
    >>>     Segregate(20, 10),
    >>>     SplitAll("label"),
    >>>     DynamicProbability2D(1.2, "v", "xy")
    >>> ])
    >>> pixels_vel = vel_pipeline.fit(trajectories)

    '''

    def __init__(
        self,
        diameter,
        column,
        dimensions = "xy",
        resolution = (512, 512),
        xlim = None,
        ylim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 2 or d[0] not in "xyz" or d[1] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly two characters containing 'x', 'y', or 'z'; e.g. "
                    f"'xy', 'zy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 2:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly two integers representing the column indices to "
                    "use; e.g. `[1, 2]` for xy, `[3, 2]` for yz. Received "
                    f"`{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]

        self.diameter = float(diameter)
        if isinstance(column, str):
            self.column = str(column)
        else:
            self.column = int(column)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        # The column whose values we will rasterize is either given as a named
        # column, or an integer index
        if isinstance(self.column, str):
            col_idx = samples.columns.index(self.column)
        else:
            col_idx = self.column

        # konigcell.probability2d will split points for parallel rasterization
        # to minimise memory consumption; need to add NaNs between points
        # samples to separate individual trajectories
        coordinates = []
        values = []
        for sample in samples:
            coordinates.append(sample.points[:, self.dimensions])
            coordinates.append(np.full(2, np.nan))

            values.append(sample.points[:, col_idx])
            values.append(np.full(1, np.nan))

        coordinates = np.vstack(coordinates)
        values = np.hstack(values)

        pixels = kc.dynamic_prob2d(
            coordinates,
            values[:-1],
            self.diameter / 2,
            resolution = self.resolution,
            xlim = self.xlim,
            ylim = self.ylim,
            max_workers = self.max_workers,
            verbose = self.verbose,
        )

        return pixels




class ResidenceDistribution2D(Reducer):
    '''Compute the 2D residence distribution of some tracer quantity (eg
    time spent in each cell).

    Reducer signature:

    ::

              PointData -> ResidenceDistribution2D.fit -> Pixels
        list[PointData] -> ResidenceDistribution2D.fit -> Pixels
          numpy.ndarray -> ResidenceDistribution2D.fit -> Pixels

    This reducer calculates the cumulative value of the tracer quantity in each
    cell of a 2D pixel grid; it uses the full projected tracer area for the
    pixelization step, so the distribution is accurate for arbitrarily fine
    resolutions.

    Parameters
    ----------
    diameter : float
        The diameter of the imaged tracer.

    column : str or int, default "t"
        The `PointData` column used to compute the residence distribution,
        given as a name (`str`) or index (`int`).

    dimensions : str or list[int], default "xy"
        The tracer coordinates used to rasterize its trajectory, given as a
        string (e.g. "xy" projects the points onto the XY plane) or a list with
        two column indices (e.g. [1, 3] for XZ).

    resolution : tuple[int, int], default (512, 512)
        The number of pixels used for the rasterization grid in the X and Y
        dimensions.

    xlim : tuple[float, float], optional
        The physical limits in the X dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    ylim : tuple[float, float], optional
        The physical limits in the y dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    max_workers : int, optional
        The maximum number of workers (threads, processes or ranks) to use by
        the parallel executor; if 1, it is sequential (and produces the
        clearest error messages should they happen). If unset, the
        ``os.cpu_count()`` is used.

    verbose : bool or str default True
        If True, time the computation and print the state of the execution.

    Examples
    --------
    Compute the residence time distribution of a single tracer trajectory:

    >>> trajectories = pept.load(...)
    >>> pixels_rtd = ResidenceDistribution2D(1.2, "t", "xy").fit(trajectories)

    Plot the pixel grid:

    >>> from pept.plots import PlotlyGrapher2D
    >>> PlotlyGrapher2D().add_pixels(pixels_rtd).show()

    For multiple tracer trajectories, you can use ``Segregate`` then
    ``SplitAll('label')`` before calling this reducer to rasterize each
    trajectory separately:

    >>> rtd_pipeline = pept.Pipeline([
    >>>     Segregate(20, 10),
    >>>     SplitAll("label"),
    >>>     ResidenceDistribution2D(1.2, "t", "xy")
    >>> ])
    >>> pixels_rtd = rtd_pipeline.fit(trajectories)

    '''

    def __init__(
        self,
        diameter,
        column = "t",
        dimensions = "xy",
        resolution = (512, 512),
        xlim = None,
        ylim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 2 or d[0] not in "xyz" or d[1] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly two characters containing 'x', 'y', or 'z'; e.g. "
                    f"'xy', 'zy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 2:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly two integers representing the column indices to "
                    "use; e.g. `[1, 2]` for xy, `[3, 2]` for yz. Received "
                    f"`{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]

        self.diameter = float(diameter)
        if isinstance(column, str):
            self.column = str(column)
        else:
            self.column = int(column)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        # The column whose values we will rasterize is either given as a named
        # column, or an integer index
        if isinstance(self.column, str):
            col_idx = samples.columns.index(self.column)
        else:
            col_idx = self.column

        # konigcell.probability2d will split points for parallel rasterization
        # to minimise memory consumption; need to add NaNs between points
        # samples to separate individual trajectories
        coordinates = []
        values = []
        for sample in samples:
            coordinates.append(sample.points[:, self.dimensions])
            coordinates.append(np.full(2, np.nan))

            values.append(sample.points[:, col_idx])
            values.append(np.full(1, np.nan))

        coordinates = np.vstack(coordinates)
        values = np.hstack(values)
        values_dt = values[1:] - values[:-1]

        pixels = kc.dynamic2d(
            coordinates,
            kc.RATIO,
            values_dt,
            self.diameter / 2,
            resolution = self.resolution,
            xlim = self.xlim,
            ylim = self.ylim,
            max_workers = self.max_workers,
            verbose = self.verbose,
        )

        return pixels




class DynamicProbability3D(Reducer):
    '''Compute the 3D probability distribution of some tracer quantity (eg
    velocity in each cell).

    Reducer signature:

    ::

              PointData -> DynamicProbability3D.fit -> Voxels
        list[PointData] -> DynamicProbability3D.fit -> Voxels
          numpy.ndarray -> DynamicProbability3D.fit -> Voxels

    This reducer calculates the average value of the tracer quantity in each
    cell of a 3D voxel grid; it uses the full projected tracer area for the
    voxelization step, so the distribution is accurate for arbitrarily fine
    resolutions.

    Parameters
    ----------
    diameter : float
        The diameter of the imaged tracer.

    column : str or int
        The `PointData` column used to compute the probability distribution,
        given as a name (`str`) or index (`int`).

    dimensions : str or list[int], default "xyz"
        The tracer coordinates used to rasterize its trajectory, given as a
        string (e.g. "xyz" or "zyx") or a list with
        three column indices (e.g. [1, 2, 3] for XYZ).

    resolution : tuple[int, int, int], default (50, 50, 50)
        The number of pixels used for the rasterization grid in the X, Y, Z
        dimensions.

    xlim : tuple[float, float], optional
        The physical limits in the X dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    ylim : tuple[float, float], optional
        The physical limits in the y dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    zlim : tuple[float, float], optional
        The physical limits in the z dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    max_workers : int, optional
        The maximum number of workers (threads, processes or ranks) to use by
        the parallel executor; if 1, it is sequential (and produces the
        clearest error messages should they happen). If unset, the
        ``os.cpu_count()`` is used.

    verbose : bool or str default True
        If True, time the computation and print the state of the execution.

    Examples
    --------
    Compute the velocity probability distribution of a single tracer trajectory
    having a column named "v" corresponding to the tracer velocity:

    >>> trajectories = pept.load(...)
    >>> voxels_vel = DynamicProbability3D(1.2, "v").fit(trajectories)

    Plot the pixel grid:

    >>> from pept.plots import PlotlyGrapher
    >>> PlotlyGrapher().add_voxels(voxels_vel).show()

    For multiple tracer trajectories, you can use ``Segregate`` then
    ``SplitAll('label')`` before calling this reducer to rasterize each
    trajectory separately:

    >>> vel_pipeline = pept.Pipeline([
    >>>     Segregate(20, 10),
    >>>     SplitAll("label"),
    >>>     DynamicProbability3D(1.2, "v")
    >>> ])
    >>> voxels_vel = vel_pipeline.fit(trajectories)

    '''

    def __init__(
        self,
        diameter,
        column,
        dimensions = "xyz",
        resolution = (50, 50, 50),
        xlim = None,
        ylim = None,
        zlim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 3 or d[0] not in "xyz" or d[1] not in "xyz" or \
                    d[2] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly three characters containing 'x', 'y', or 'z'; "
                    f"e.g. 'xyz', 'zxy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
            self.dimensions[2] = ord(d[2]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 3:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly three integers representing the column indices "
                    "to use; e.g. `[1, 2, 3]` for xyz, `[3, 1, 2]` for zxy. "
                    f"Received `{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]
            self.dimensions[2] = d[2]

        self.diameter = float(diameter)
        if isinstance(column, str):
            self.column = str(column)
        else:
            self.column = int(column)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim
        self.zlim = zlim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        # The column whose values we will rasterize is either given as a named
        # column, or an integer index
        if isinstance(self.column, str):
            col_idx = samples.columns.index(self.column)
        else:
            col_idx = self.column

        # konigcell.probability2d will split points for parallel rasterization
        # to minimise memory consumption; need to add NaNs between points
        # samples to separate individual trajectories
        coordinates = []
        values = []
        for sample in samples:
            coordinates.append(sample.points[:, self.dimensions])
            coordinates.append(np.full(3, np.nan))

            values.append(sample.points[:, col_idx])
            values.append(np.full(1, np.nan))

        coordinates = np.vstack(coordinates)
        values = np.hstack(values)

        voxels = kc.dynamic_prob3d(
            coordinates,
            values[:-1],
            self.diameter / 2,
            resolution = self.resolution,
            xlim = self.xlim,
            ylim = self.ylim,
            zlim = self.zlim,
            max_workers = self.max_workers,
            verbose = self.verbose,
        )

        return voxels




class ResidenceDistribution3D(Reducer):
    '''Compute the 3D residence distribution of some tracer quantity (eg
    time spent in each cell).

    Reducer signature:

    ::

              PointData -> ResidenceDistribution3D.fit -> Pixels
        list[PointData] -> ResidenceDistribution3D.fit -> Pixels
          numpy.ndarray -> ResidenceDistribution3D.fit -> Pixels

    This reducer calculates the cumulative value of the tracer quantity in each
    cell of a 3D voxel grid; it uses the full projected tracer area for the
    voxelization step, so the distribution is accurate for arbitrarily fine
    resolutions.

    Parameters
    ----------
    diameter : float
        The diameter of the imaged tracer.

    column : str or int
        The `PointData` column used to compute the probability distribution,
        given as a name (`str`) or index (`int`).

    dimensions : str or list[int], default "xyz"
        The tracer coordinates used to rasterize its trajectory, given as a
        string (e.g. "xyz" or "zyx") or a list with
        three column indices (e.g. [1, 2, 3] for XYZ).

    resolution : tuple[int, int, int], default (50, 50, 50)
        The number of pixels used for the rasterization grid in the X, Y, Z
        dimensions.

    xlim : tuple[float, float], optional
        The physical limits in the X dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    ylim : tuple[float, float], optional
        The physical limits in the y dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    zlim : tuple[float, float], optional
        The physical limits in the z dimension of the pixel grid. If unset, it
        is automatically computed to contain all tracer positions (default).

    max_workers : int, optional
        The maximum number of workers (threads, processes or ranks) to use by
        the parallel executor; if 1, it is sequential (and produces the
        clearest error messages should they happen). If unset, the
        ``os.cpu_count()`` is used.

    verbose : bool or str default True
        If True, time the computation and print the state of the execution.

    Examples
    --------
    Compute the residence time distribution of a single tracer trajectory:

    >>> trajectories = pept.load(...)
    >>> voxels_rtd = ResidenceDistribution3D(1.2, "t").fit(trajectories)

    Plot the pixel grid:

    >>> from pept.plots import PlotlyGrapher
    >>> PlotlyGrapher().add_voxels(voxels_rtd).show()

    For multiple tracer trajectories, you can use ``Segregate`` then
    ``SplitAll('label')`` before calling this reducer to rasterize each
    trajectory separately:

    >>> rtd_pipeline = pept.Pipeline([
    >>>     Segregate(20, 10),
    >>>     SplitAll("label"),
    >>>     ResidenceDistribution3D(1.2, "t")
    >>> ])
    >>> voxels_rtd = rtd_pipeline.fit(trajectories)

    '''

    def __init__(
        self,
        diameter,
        column = "t",
        dimensions = "xyz",
        resolution = (50, 50, 50),
        xlim = None,
        ylim = None,
        zlim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 3 or d[0] not in "xyz" or d[1] not in "xyz" or \
                    d[2] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly three characters containing 'x', 'y', or 'z'; "
                    f"e.g. 'xyz', 'zxy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
            self.dimensions[2] = ord(d[2]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 3:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly three integers representing the column indices "
                    "to use; e.g. `[1, 2, 3]` for xyz, `[3, 1, 2]` for zxy. "
                    f"Received `{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]
            self.dimensions[2] = d[2]

        self.diameter = float(diameter)
        if isinstance(column, str):
            self.column = str(column)
        else:
            self.column = int(column)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim
        self.zlim = zlim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):
        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        # The column whose values we will rasterize is either given as a named
        # column, or an integer index
        if isinstance(self.column, str):
            col_idx = samples.columns.index(self.column)
        else:
            col_idx = self.column

        # konigcell.probability2d will split points for parallel rasterization
        # to minimise memory consumption; need to add NaNs between points
        # samples to separate individual trajectories
        coordinates = []
        values = []
        for sample in samples:
            coordinates.append(sample.points[:, self.dimensions])
            coordinates.append(np.full(3, np.nan))

            values.append(sample.points[:, col_idx])
            values.append(np.full(1, np.nan))

        coordinates = np.vstack(coordinates)
        values = np.hstack(values)
        values_dt = values[1:] - values[:-1]

        voxels = kc.dynamic3d(
            coordinates,
            kc.RATIO,
            values_dt,
            self.diameter / 2,
            resolution = self.resolution,
            xlim = self.xlim,
            ylim = self.ylim,
            zlim = self.zlim,
            max_workers = self.max_workers,
            verbose = self.verbose,
        )

        return voxels




class VectorField2D(Reducer):
    '''Compute a 2D vector field - effectively two 2D grids computed from
    two columns, for example X and Y velocities.

    Reducer signature:

    ::

              PointData -> VectorField2D.fit -> VectorGrid2D
        list[PointData] -> VectorField2D.fit -> VectorGrid2D
          numpy.ndarray -> VectorField2D.fit -> VectorGrid2D

    Examples
    --------
    Compute a velocity vector field in the Y and Z dimensions (velocities
    were first calculated using ``pept.tracking.Velocity``):

    >>> from pept.processing import *
    >>> trajectories = pept.PointData(...)
    >>> field = VectorField2D(0.6, ["vy", "vz"], "yz").fit(trajectories)
    >>> field
    VectorGrid2D(xpixels, ypixels)

    Create a quiver plot using Plotly (may be a bit slow):

    >>> scaling = 16
    >>> fig = field.quiver(scaling)
    >>> fig.show()

    Create a 2D vector field (needs PyVista):

    >>> scaling = 16
    >>> fig = field.vectors(scaling)
    >>> fig.plot(cmap = "magma")

    '''

    def __init__(
        self,
        diameter,
        columns = ["vx", "vy"],
        dimensions = "xy",
        resolution = (50, 50),
        xlim = None,
        ylim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 2 or d[0] not in "xyz" or d[1] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly two characters containing 'x' or 'y'; "
                    f"e.g. 'xy', 'zy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 2:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly two integers representing the column indices "
                    "to use; e.g. `[1, 2]` for xy, `[3, 1]` for zx. "
                    f"Received `{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]

        self.diameter = float(diameter)

        columns = list(columns)
        if len(columns) != 2:
            raise ValueError(textwrap.fill((
                "The input `columns` must be a list-like with exactly two "
                "elements, either column names (str, e.g. 'vx') or column "
                f"indices (int, e.g. 4). Received {columns}."
            )))

        self.columns = [None, None]
        for i, col in enumerate(columns):
            if isinstance(col, str):
                self.columns[i] = str(col)
            else:
                self.columns[i] = int(col)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):
        if self.verbose:
            start = time.time()

        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        grids = [None, None]
        for i, col in enumerate(self.columns):
            if self.verbose:
                print(f"Step {i + 1} / {len(self.columns)}:")

            pixelizer = DynamicProbability2D(
                self.diameter,
                col,
                self.dimensions,
                self.resolution,
                self.xlim,
                self.ylim,
                max_workers = self.max_workers,
                verbose = self.verbose,
            )
            grids[i] = pixelizer.fit(samples)

        if self.verbose:
            end = time.time()
            print(f"Compute 3D vector field in {end - start:4.4f} s.")

        return VectorGrid2D(*grids)




class VectorGrid2D:
    '''Object produced by ``VectorField2D`` storing 2 grids of voxels
    `xpixels`, `ypixels`, for example velocity vector fields / quiver plots.

    Examples
    --------
    Compute a velocity vector field in the Y and Z dimensions (velocities
    were first calculated using ``pept.tracking.Velocity``):

    >>> from pept.processing import *
    >>> trajectories = pept.PointData(...)
    >>> field = VectorField2D(0.6, ["vy", "vz"], "yz").fit(trajectories)
    >>> field
    VectorGrid2D(xpixels, ypixels)

    Create a quiver plot using Plotly (may be a bit slow):

    >>> scaling = 16
    >>> fig = field.quiver(scaling)
    >>> fig.show()

    Create a 2D vector field (needs PyVista):

    >>> scaling = 16
    >>> fig = field.vectors(scaling)
    >>> fig.plot(cmap = "magma")

    '''

    def __init__(
        self,
        xpixels: Pixels,
        ypixels: Pixels,
    ):
        self.xpixels = xpixels
        self.ypixels = ypixels


    def vectors(self, factor = 1):
        # You need to install PyVista to use this function!
        grid = pv.UniformGrid()

        shape = self.xpixels.pixels.shape
        grid.dimensions = [shape[0] + 1, shape[1] + 1, 1]
        grid.origin = list(self.xpixels.lower) + [0]
        grid.spacing = list(self.xpixels.pixel_size) + [0]
        grid.cell_data.values = self.xpixels.pixels.flatten(order="F")

        vectors = grid.cell_centers()

        vectors["vectors"] = np.vstack((
            self.xpixels.pixels.flatten(order="F"),
            self.ypixels.pixels.flatten(order="F"),
            np.zeros(len(self.xpixels.pixels.flatten())),
        )).T

        vectors.active_vectors_name = 'vectors'
        return vectors.glyph(factor = factor)


    def quiver(self, factor = 1):
        xg, yg = self.xpixels.pixel_grids
        xg = 0.5 * (xg[1:] + xg[:-1])
        yg = 0.5 * (yg[1:] + yg[:-1])

        x, y = np.meshgrid(xg, yg)
        fig = ff.create_quiver(
            x,
            y,
            self.xpixels.pixels.T,
            self.ypixels.pixels.T,
            scale = factor,
        )
        return fig


    def __repr__(self):
        return "VectorGrid2D(xpixels, ypixels)"




class VectorField3D(Reducer):
    '''Compute a 3D vector field - effectively three 3D grids computed from
    three columns, for example X, Y and Z velocities.

    Reducer signature:

    ::

              PointData -> VectorField3D.fit -> VectorGrid3D
        list[PointData] -> VectorField3D.fit -> VectorGrid3D
          numpy.ndarray -> VectorField3D.fit -> VectorGrid3D

    Examples
    --------
    Compute a 3D velocity vector field (velocities were first calculated using
    ``pept.tracking.Velocity``):

    >>> from pept.processing import *
    >>> trajectories = pept.PointData(...)
    >>> field = VectorField3D(0.6).fit(trajectories)
    >>> field
    VectorGrid3D(xvoxels, yvoxels, zvoxels)

    Create a 3D vector field (needs PyVista):

    >>> scaling = 16
    >>> fig = field.vectors(scaling)
    >>> fig.plot(cmap = "magma")

    '''

    def __init__(
        self,
        diameter,
        columns = ["vx", "vy", "vz"],
        dimensions = "xyz",
        resolution = (50, 50, 50),
        xlim = None,
        ylim = None,
        zlim = None,
        max_workers = None,
        verbose = True,
    ):
        # Type-checking inputs
        self.dimensions = [None, None, None]
        if isinstance(dimensions, str):
            d = str(dimensions)
            if len(d) != 3 or d[0] not in "xyz" or d[1] not in "xyz" or \
                    d[2] not in "xyz":
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a str, must have "
                    "exactly three characters containing 'x', 'y', or 'z'; "
                    f"e.g. 'xyz', 'zxy'. Received `{d}`."
                )))

            # Transform x -> 1, y -> 2, z -> 3 using ASCII integer `ord`er
            self.dimensions[0] = ord(d[0]) - ord('x') + 1
            self.dimensions[1] = ord(d[1]) - ord('x') + 1
            self.dimensions[2] = ord(d[2]) - ord('x') + 1
        else:
            d = np.asarray(dimensions, dtype = int)
            if d.ndim != 1 or len(d) != 3:
                raise ValueError(textwrap.fill((
                    "The input `dimensions`, if given as a list, must contain "
                    "exactly three integers representing the column indices "
                    "to use; e.g. `[1, 2, 3]` for xyz, `[3, 1, 2]` for zxy. "
                    f"Received `{d}`."
                )))
            self.dimensions[0] = d[0]
            self.dimensions[1] = d[1]
            self.dimensions[2] = d[2]

        self.diameter = float(diameter)

        columns = list(columns)
        if len(columns) != 3:
            raise ValueError(textwrap.fill((
                "The input `columns` must be a list-like with exactly 3 "
                "elements, either column names (str, e.g. 'vx') or column "
                f"indices (int, e.g. 4). Received {columns}."
            )))

        self.columns = [None, None, None]
        for i, col in enumerate(columns):
            if isinstance(col, str):
                self.columns[i] = str(col)
            else:
                self.columns[i] = int(col)

        self.resolution = resolution
        self.xlim = xlim
        self.ylim = ylim
        self.zlim = zlim

        self.max_workers = max_workers
        self.verbose = verbose


    def fit(self, samples):

        if self.verbose:
            start = time.time()

        # Reduce / stack list of samples onto a single PointData / array
        samples = Stack().fit(samples)

        if not isinstance(samples, PointData):
            samples = PointData(samples)

        grids = [None, None, None]
        for i, col in enumerate(self.columns):
            if self.verbose:
                print(f"Step {i + 1} / {len(self.columns)}...")

            voxelizer = DynamicProbability3D(
                self.diameter,
                col,
                self.dimensions,
                self.resolution,
                self.xlim,
                self.ylim,
                self.zlim,
                max_workers = self.max_workers,
                verbose = self.verbose,
            )
            grids[i] = voxelizer.fit(samples)

        if self.verbose:
            end = time.time()
            print(f"Compute 3D vector field in {end - start:4.4f} s.")

        return VectorGrid3D(*grids)




class VectorGrid3D:
    '''Object produced by ``VectorField3D`` storing 3 grids of voxels
    `xvoxels`, `yvoxels`, `zvoxels`, for example velocity vector fields /
    quiver plots.

    Examples
    --------
    Compute a 3D velocity vector field (velocities were first calculated using
    ``pept.tracking.Velocity``):

    >>> from pept.processing import *
    >>> trajectories = pept.PointData(...)
    >>> field = VectorField3D(0.6).fit(trajectories)
    >>> field
    VectorGrid3D(xvoxels, yvoxels, zvoxels)

    Create a 3D vector field (needs PyVista):

    >>> scaling = 16
    >>> fig = field.vectors(scaling)
    >>> fig.plot(cmap = "magma")

    '''

    def __init__(
        self,
        xvoxels: Voxels,
        yvoxels: Voxels,
        zvoxels: Voxels,
    ):
        self.xvoxels = xvoxels
        self.yvoxels = yvoxels
        self.zvoxels = zvoxels


    def vectors(self, factor = 1):
        # You need to install PyVista to use this function!
        grid = pv.UniformGrid()
        grid.dimensions = np.array(self.xvoxels.voxels.shape) + 1
        grid.origin = self.xvoxels.lower
        grid.spacing = self.xvoxels.voxel_size
        grid.cell_data.values = self.xvoxels.voxels.flatten(order="F")

        vectors = grid.cell_centers()

        vectors["vectors"] = np.vstack((
            self.xvoxels.voxels.flatten(order="F"),
            self.yvoxels.voxels.flatten(order="F"),
            self.zvoxels.voxels.flatten(order="F"),
        )).T

        vectors.active_vectors_name = 'vectors'
        return vectors.glyph(factor = factor)


    def __repr__(self):
        return "VectorGrid3D(xvoxels, yvoxels, zvoxels)"
